//--------------------------------------------------------------------------------------
// File: SIMD.cpp
//--------------------------------------------------------------------------------------

#include "SIMD.h"
#include "rgbe.h"
#include "timer.h"
#include "stdafx.h"
#include <stdio.h>

//global timer
CTimer g_myTimer;
__int64 g_start, g_stop;
__int64 g_speed = g_myTimer.GetCPUSpeed();

//global variables
extern bool					g_bSIMDConvert;
extern double				g_dLumCalcTime;
extern float				g_fCurrentImageKey;
extern bool					g_firstrun;
extern float				g_fPreviousImageKey ;
extern float				g_fImageKey;

//--------------------------------------------------------------------------------------
// Name: CreateRGBETextureFromFile()
//--------------------------------------------------------------------------------------

HRESULT CreateRGBETextureFromRLEFile( IDirect3DDevice9* pd3dDevice, LPDIRECT3DTEXTURE9* ppTexDest_RGBE8, char* strHDRFile )
{
	FLOAT r, g, b;
	int width, height;

	FILE* pHDRFile;
	if ( (pHDRFile = fopen( strHDRFile, "rb" )) == NULL )
	{
		return S_FALSE;
	}

	rgbe_header_info pInfo;
	RGBE_ReadHeader( pHDRFile, &width, &height, &pInfo );

	//3 color channels per pixel
	float* pHDR = new float[3 * width * height];

	memset( pHDR, 0, 3 * width * height * sizeof( float ) );

	//reads RGBE RLE data, creates 3 HDR float results (1 per channel)
    RGBE_ReadPixels_RLE( pHDRFile, pHDR, width, height );

	//Convert it to unencoded RGBE
	BYTE* pRGBE = new BYTE[4 * width * height];

	int i = 0, j = 0;

#ifdef SIMD_RGBE_CONVERT	//defined in SIMD.h

	g_bSIMDConvert = true;

	for( i, j; i < (3 * (width * height)) - 3; i += 12, j += 1 )
	{
		U_128 max128, fDivisor;
	    
		#if 0  //sets
			//load 4 floats per channel
			__m128 r128 = _mm_set_ps ((FLOAT) *(pHDR + i + 9), (FLOAT) *(pHDR + i + 6),  (FLOAT) *(pHDR + i + 3), (FLOAT) *(pHDR + i + 0));
			__m128 g128 = _mm_set_ps ((FLOAT) *(pHDR + i + 10), (FLOAT) *(pHDR + i + 7),  (FLOAT) *(pHDR + i + 4), (FLOAT) *(pHDR + i + 1));
			__m128 b128 = _mm_set_ps ((FLOAT) *(pHDR + i + 11), (FLOAT) *(pHDR + i + 8), (FLOAT) *(pHDR + i + 5), (FLOAT) *(pHDR + i + 2));
		#else //swizzle (faster)
			__m128 load0_128 = _mm_loadu_ps ( (const float *)(pHDR + i + 0) );  // r1 | b0 | g0 | r0
			__m128 load1_128 = _mm_loadu_ps ( (const float *)(pHDR + i + 4) );  // g2 | r2 | b1 | g1
			__m128 load2_128 = _mm_loadu_ps ( (const float *)(pHDR + i + 8) );  // b3 | g3 | r3 | b2

			__m128 rg_128 = _mm_shuffle_ps ( load0_128, load1_128, 204 );	// g2 | g1 | r1 | r0
			__m128 br_128 = _mm_shuffle_ps ( load1_128, load2_128, 73 );	// r3 | b2 | r2 | b1
			__m128 gb_128 = _mm_shuffle_ps ( load2_128, load0_128, 110 );	// g0 | b0 | b3 | g3

			__m128 r128 = _mm_shuffle_ps ( rg_128, br_128, 212 );			// r3 | r2 | r1 | r0

			__m128 pre_g128 = _mm_shuffle_ps ( gb_128, rg_128, 227 );		// g2 | g1 | g3 | g0
			__m128 g128 = _mm_shuffle_ps ( pre_g128, pre_g128, 120 );		// g3 | g2 | g1 | g0

			__m128 pre_b128 = _mm_shuffle_ps ( gb_128, br_128, 134 );		// b2 | b1 | b3 | b0
			__m128 b128 = _mm_shuffle_ps ( pre_b128, pre_b128, 120 );		// b3 | b2 | b1 | b0
		#endif


		//grab max
		max128.f = _mm_max_ps ( b128, _mm_max_ps ( g128, r128 ) );
	    
		//exploit floating point format
		__m128i nExp128 = _mm_sub_epi32( _mm_srli_epi32 ( max128.i, 23 ), onetwentyseven_128i );
		__m128i tempnExp128 = _mm_add_epi32 ( nExp128, one_128i );
		__m128i nadd128 = _mm_and_si128( max128.i, oo7ffff_128i );
		__m128i mask128 = _mm_cmpgt_epi32( nadd128, one_128i );
		__m128i replace128 = _mm_and_si128( mask128, tempnExp128 );
		__m128i keep128 = _mm_andnot_si128( mask128, nExp128 );
		nExp128 = _mm_or_si128( keep128, replace128 );

		nExp128 = _mm_add_epi32 ( nExp128, onetwentyeight_128i );  //permanent re-bias since we use later
		fDivisor.i = _mm_or_si128( _mm_slli_epi32( nExp128, 23 ), one_128i );

		r128 = _mm_div_ps ( r128, fDivisor.f );
		g128 = _mm_div_ps ( g128, fDivisor.f );
		b128 = _mm_div_ps ( b128, fDivisor.f );

		r128 = _mm_max_ps( zero_128, _mm_min_ps( one_128, r128 ) );
		g128 = _mm_max_ps( zero_128, _mm_min_ps( one_128, g128 ) );
		b128 = _mm_max_ps( zero_128, _mm_min_ps( one_128, b128 ) );
	
		r128 = _mm_mul_ps( r128, twofiftyfive_128 );
		g128 = _mm_mul_ps( g128, twofiftyfive_128 );
		b128 = _mm_mul_ps( b128, twofiftyfive_128 );

		__m128i r128i = _mm_cvtps_epi32( r128 );
		__m128i g128i = _mm_cvtps_epi32( g128 );
		__m128i b128i = _mm_cvtps_epi32( b128 );

		nExp128 = _mm_slli_epi32( nExp128, 24 );
		r128i   = _mm_slli_epi32( r128i, 16 );
		g128i   = _mm_slli_epi32( g128i, 8 );

		__m128i pixelsout = _mm_or_si128( nExp128, _mm_or_si128( r128i, _mm_or_si128( g128i, b128i ) ) );
		_mm_storeu_si128( (__m128i*)(pRGBE + j * 16), pixelsout );
		
	}


	//edge case
	for( i, j; i < 3 * (width * height); i += 3, j += 4 )
	{

		r = pHDR[i];
		g = pHDR[i + 1];
		b = pHDR[i + 2];

		// Determine the largest color component
		float maxComponent = max( max(r, g), b );
	    
		//Round to the nearest integer exponent
		int nmax = *(int *)&maxComponent;
		int nExp = (nmax >> 23) - 127;						//put exponent in bottom and unbias it
		int nadd = nmax & 0x007ffff;						//get mantissa bits (to determine if power of 2)
		if (nadd)
			nExp++;
		int nDivisor = (((nExp + 128) << 23) | (0x1));		//re-bias the exponent
		float fDivisor = *(float *)&nDivisor;				//generate the float it represents

		r /= fDivisor;
		g /= fDivisor;
		b /= fDivisor;

		// Constrain the color components
		r = max( 0, min(1, r) );
		g = max( 0, min(1, g) );
		b = max( 0, min(1, b) );

		// Store the shared exponent in the alpha channel
		pRGBE[j + 0] = (BYTE)(b*255);
		pRGBE[j + 1] = (BYTE)(g*255);
		pRGBE[j + 2] = (BYTE)(r*255);
		pRGBE[j + 3] = (BYTE)(nExp+128);		
	}
	
#else

	for( i, j; i < 3 * (width * height); i += 3, j += 4 )
		{

		r = pHDR[i];
		g = pHDR[i + 1];
		b = pHDR[i + 2];

		// Determine the largest color component
		float maxComponent = max( max(r, g), b );
	    
		//Round to the nearest integer exponent
		int nmax = *(int *)&maxComponent;
		int nExp = (nmax >> 23) - 127;						//put exponent in bottom and unbias it
		int nadd = nmax & 0x007ffff;						//get mantissa bits (to determine if power of 2)
		if (nadd)
			nExp++;
		int nDivisor = (((nExp + 128) << 23) | (0x1));		//re-bias the exponent
		float fDivisor = *(float *)&nDivisor;				//generate the float it represents

		r /= fDivisor;
		g /= fDivisor;
		b /= fDivisor;

		// Constrain the color components
		r = max( 0, min(1, r) );
		g = max( 0, min(1, g) );
		b = max( 0, min(1, b) );

		// Store the shared exponent in the alpha channel
		pRGBE[j + 0] = (BYTE)(b*255);
		pRGBE[j + 1] = (BYTE)(g*255);
		pRGBE[j + 2] = (BYTE)(r*255);
		pRGBE[j + 3] = (BYTE)(nExp+128);		
	}
#endif

	HRESULT hr;
	hr = D3DXCreateTexture( pd3dDevice, width, height, 1, D3DUSAGE_DYNAMIC, 
								   D3DFMT_A8R8G8B8, D3DPOOL_SYSTEMMEM, ppTexDest_RGBE8 );
	if FAILED( hr ) {
		return( hr );
	}	

	D3DLOCKED_RECT pLR;
	(*ppTexDest_RGBE8)->LockRect( 0, &pLR, NULL, 0 );

	memcpy( pLR.pBits, pRGBE, width * height * sizeof( DWORD ) );

	(*ppTexDest_RGBE8)->UnlockRect( 0 );

	delete[] pHDR;
	delete[] pHDRFile;

	return S_OK;

}


//-----------------------------------------------------------------------------
//  MeasureLuminance()
//-----------------------------------------------------------------------------

HRESULT MeasureLuminance(IDirect3DDevice9* pd3dDevice, LPDIRECT3DSURFACE9 pSourceRenderTarget, LPDIRECT3DSURFACE9 g_pLumSurf, bool g_bUseSIMD) 
{
	HRESULT hr = S_OK;
	DWORD RGBEpixel;
	int Exp, ShiftRes;	
	int i = 0, j = 0;
	float Rfloat, Gfloat, Bfloat;
	float fExp = 0, fExp2 = 0, PixelLuminance = 0, TotalLuminance = 0, number_of_pixels = 0;
	U_128 Exp_128;
	__m128 TotalLum_128 = _mm_setzero_ps();

	LPDIRECT3DSURFACE9 pPixelData = NULL;
	pPixelData = g_pLumSurf;
	V_RETURN( pd3dDevice->GetRenderTargetData( pSourceRenderTarget, pPixelData ) );

	D3DSURFACE_DESC descSrc;
    pPixelData->GetDesc( &descSrc );

	number_of_pixels = 	descSrc.Height * descSrc.Width;

	// Lock the source texture for reading
    D3DLOCKED_RECT rcSrc;
    V_RETURN( pPixelData->LockRect( &rcSrc, NULL, D3DLOCK_READONLY ) );

	//location of texture data
	BYTE* pSourcePixels = (BYTE *)rcSrc.pBits;	

	
	int localwidth = descSrc.Width;
	int localheight = descSrc.Height; 	
	g_start = g_myTimer.GetTime();
	
	//Get full scene luminance
	if( g_bUseSIMD )
	{
		for( i = 0; i < localheight; i++ )
		{
			for( j = 0; j < (localwidth-3); j+=4 )
			{
				RGBEpixel = i * (localwidth) + j;
				__m128i RGBE0_128 = _mm_loadu_si128((__m128i*)(pSourcePixels + 4 * RGBEpixel));
				__m128i RGBE1_128 = _mm_srli_si128( RGBE0_128, 4);
				__m128i RGBE2_128 = _mm_srli_si128( RGBE0_128, 8);
				__m128i RGBE3_128 = _mm_srli_si128( RGBE0_128, 12);

				__m128i RGBE_WORD0_128	= _mm_unpacklo_epi8( RGBE0_128, zero_128i );
				__m128i RGBE_DWORD0_128	= _mm_unpacklo_epi16( RGBE_WORD0_128, zero_128i );  // E0 | B0 | G0 | R0 
	            
				__m128i RGBE_WORD1_128	= _mm_unpacklo_epi8( RGBE1_128, zero_128i );
				__m128i RGBE_DWORD1_128	= _mm_unpacklo_epi16( RGBE_WORD1_128, zero_128i );  // E1 | B1 | G1 | R1
				
				__m128i RGBE_WORD2_128	= _mm_unpacklo_epi8( RGBE2_128, zero_128i );
				__m128i RGBE_DWORD2_128	= _mm_unpacklo_epi16( RGBE_WORD2_128, zero_128i );  // E2 | B2 | G2 | R2
				
				__m128i RGBE_WORD3_128	= _mm_unpacklo_epi8( RGBE3_128, zero_128i );
				__m128i RGBE_DWORD3_128	= _mm_unpacklo_epi16( RGBE_WORD3_128, zero_128i );  // E3 | B3 | G3 | R3 
				
				__m128i R0R1G0G1_128 = _mm_unpacklo_epi32( RGBE_DWORD0_128, RGBE_DWORD1_128 );
				__m128i R2R3G2G3_128 = _mm_unpacklo_epi32( RGBE_DWORD2_128, RGBE_DWORD3_128 );
				__m128i B0B1E0E1_128 = _mm_unpackhi_epi32( RGBE_DWORD0_128, RGBE_DWORD1_128 );
				__m128i B2B3E2E3_128 = _mm_unpackhi_epi32( RGBE_DWORD2_128, RGBE_DWORD3_128 );

				__m128i R0R1R2R3_128 = _mm_unpacklo_epi64( R0R1G0G1_128, R2R3G2G3_128 );
				__m128i G0G1G2G3_128 = _mm_unpackhi_epi64( R0R1G0G1_128, R2R3G2G3_128 );
				__m128i B0B1B2B3_128 = _mm_unpacklo_epi64( B0B1E0E1_128, B2B3E2E3_128 );
				__m128i E0E1E2E3_128 = _mm_unpackhi_epi64( B0B1E0E1_128, B2B3E2E3_128 );

				__m128  Rfloat_128	= _mm_cvtepi32_ps( R0R1R2R3_128 );
				__m128  Gfloat_128	= _mm_cvtepi32_ps( G0G1G2G3_128 );
				__m128  Bfloat_128	= _mm_cvtepi32_ps( B0B1B2B3_128 );

				//ShiftRes = (Exp) << 23;
				//fExp2 = *(float *)&ShiftRes;			
				Exp_128.i	= _mm_slli_epi32( E0E1E2E3_128, 23 );
				                       
				__m128 pixelR_128	= _mm_mul_ps( Exp_128.f, _mm_mul_ps( Rfloat_128, oneover255_128 ) );
				__m128 pixelG_128	= _mm_mul_ps( Exp_128.f, _mm_mul_ps( Gfloat_128, oneover255_128 ) );
				__m128 pixelB_128	= _mm_mul_ps( Exp_128.f, _mm_mul_ps( Bfloat_128, oneover255_128 ) );

				__m128 LumR_128	= _mm_mul_ps( pixelR_128, point2125_128 );
				__m128 LumG_128	= _mm_mul_ps( pixelG_128, point7154_128 );
				__m128 LumB_128	= _mm_mul_ps( pixelB_128, point0721_128 );

				__m128 PixelLum_128	= _mm_add_ps( point001_128, _mm_add_ps( LumB_128, _mm_add_ps( LumG_128, LumR_128 ) ) );
				__m128 PixelLumShuf0_128 = _mm_shuffle_ps( PixelLum_128, PixelLum_128, 14);
				__m128 PixelLumSum0_128	= _mm_add_ps( PixelLum_128, PixelLumShuf0_128 );
				__m128 PixelLumShuf1_128 = _mm_shuffle_ps( PixelLumSum0_128, PixelLumSum0_128, 1);
				__m128 PixelLumSum1_128	= _mm_add_ps( PixelLumShuf1_128, PixelLumSum0_128 );

				TotalLum_128 = _mm_add_ps( TotalLum_128, PixelLumSum1_128 );
			}
		}

		_mm_store_ss ( (float *)&TotalLuminance, TotalLum_128 );

		for( i ; i < localheight; i++ )
		{
			for( j; j < localwidth; j++ )
			{
				RGBEpixel = i * (localwidth) + j;
				Exp = (pSourcePixels[4 * RGBEpixel + 3]);
				ShiftRes = (Exp) << 23;
				fExp2 = *(float *)&ShiftRes;
				Rfloat = ((float)pSourcePixels[4 * RGBEpixel + 0]/255) * fExp2;
				Gfloat = ((float)pSourcePixels[4 * RGBEpixel + 1]/255) * fExp2;
				Bfloat = ((float)pSourcePixels[4 * RGBEpixel + 2]/255) * fExp2;

				PixelLuminance = 0.2125 * Rfloat + 0.7154 * Gfloat + 0.0721 * Bfloat;

				TotalLuminance += ((PixelLuminance + 0.001));
			}
		}
	}
	else
	{
		for( i = 0; i < localheight; i++ )
		{
			for( j = 0; j < localwidth; j++ )
			{
				RGBEpixel = i * (localwidth) + j;
				Exp = (pSourcePixels[4 * RGBEpixel + 3]);
				ShiftRes = (Exp) << 23;
				fExp2 = *(float *)&ShiftRes;
				Rfloat = ((float)pSourcePixels[4 * RGBEpixel + 0]/255) * fExp2;
				Gfloat = ((float)pSourcePixels[4 * RGBEpixel + 1]/255) * fExp2;
				Bfloat = ((float)pSourcePixels[4 * RGBEpixel + 2]/255) * fExp2;

				PixelLuminance = 0.2125 * Rfloat + 0.7154 * Gfloat + 0.0721 * Bfloat;

				TotalLuminance += ((PixelLuminance + 0.001));
			}
		}
	}

	g_stop = g_myTimer.GetTime();

	g_dLumCalcTime = (double)g_stop - (double)g_start;	

	pPixelData->UnlockRect();	

	g_fCurrentImageKey = (1/number_of_pixels) * TotalLuminance;

	if(g_firstrun)
	{
		g_fPreviousImageKey = g_fCurrentImageKey;
		g_firstrun = !g_firstrun;
	}

	float diff = (g_fCurrentImageKey - g_fPreviousImageKey);

	//constrain luminance change to be more realistic
	if (diff > 0.05)
	{
		g_fImageKey = g_fPreviousImageKey + 0.05;
		g_fPreviousImageKey = g_fImageKey;
	}
	else if (diff < -0.05)
	{
		g_fImageKey = g_fPreviousImageKey - 0.05;
		g_fPreviousImageKey = g_fImageKey;
	}
	else
	{
		g_fImageKey = g_fCurrentImageKey;
		g_fPreviousImageKey = g_fImageKey;
	}
	
	return S_OK;
}